import java.util.ArrayList;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.NoSuchElementException;
import java.util.Stack;

public class BinarySearchTree<T extends Comparable<? super T>> {
	private BinaryNode root;
	private int size;
	
	public BinarySearchTree(){
		root = null;
		size = 0;
	}

	/** 
	 * This procedure reverses a BST, so that it is now ordered from largest
	 * to smallest element. This procedure produces a tree that is symmetric to
	 * the old tree. For example, if the current tree is as shown on the left side,
	 * then this method will turn that tree into the one shown on the right side.
	 * 
	 *                   15                        15
	 *                 /    \                    /    \
	 *                10    20                 20     10
	 *                  \                            /
	 *                  12                          12
	 * 
	 * This method runs in linear time. You may not insert the elements into a
	 * new tree. 
	 */
	public void reverse() {
	}
	
	public class MyInteger {
		private int value;
		public MyInteger(int v) {
			this.value = v;
		}
	}
	
	/** 
	 * This procedure determines whether a BST in fact satisfies the requirements of
	 * an AVL tree. A BST satisfies this requirement if for every node in the tree
	 * the difference in height of its children is at most one.
	 * This method runs in linear time. You may not add a height field to the BST or
	 * Node classes. You may not use a height method either.
	 *  
	 * @return This method returns true, if the BST forms an AVL tree and false otherwise.
	 */
	public boolean isAVLTree(){
		return false;
	}

	public boolean insert(T element) {
		if (element == null) throw new IllegalArgumentException();
		if (root == null) {
			root = new BinaryNode(element);
			size++;
			return true;
		}
		return root.insert(element);
	}
	
	public int size() {
		return this.size;
	}
	
	public boolean isEmpty() {
		return root == null;
	}
	
	public Iterator<BinaryNode> iterator(){
		return new TreeIterator();
	}
	
	public Iterator<BinarySearchTree<T>.BinaryNode> preOrderIterator(){
		return new PreOrderIterator();
	}
	
	
	public class BinaryNode {
		private T element;
		private BinaryNode leftChild;
		private BinaryNode rightChild;
		
		public BinaryNode(T element){
			this.element = element;
			this.leftChild = null;
			this.rightChild = null;		
		}
		
		public T getElement() {
			return element;
		}
		
		public BinaryNode getLeftChild() {
			return leftChild;
		}
		public BinaryNode getRightChild() {
			return rightChild;
		}
		
		public boolean insert(T element) {
			int comp = element.compareTo(this.element);
			if (comp < 0) {
				if (leftChild == null) {
					leftChild = new BinaryNode(element);
					size++;
					return true;
				}
				return leftChild.insert(element);
			}
			if (comp > 0) {
				if (rightChild == null) {
					rightChild = new BinaryNode(element);
					size++;
					return true;
				}
				return rightChild.insert(element);
			}
			return false;
		}
		
		public ArrayList<BinaryNode> getLeftAncestors(){
			ArrayList<BinaryNode> temp = new ArrayList<BinaryNode>();
			BinaryNode ptr = this;
			while (ptr != null) {
				temp.add(0,ptr);
				ptr = ptr.leftChild;
			}
			return temp;
		}
	}

	
	// This is an in-order iterator
	private class TreeIterator implements Iterator<BinaryNode> {
		private LinkedList<BinaryNode> nodes;

		public TreeIterator(){
			nodes = new LinkedList<BinaryNode>();
			if (root != null){
				nodes.addAll(root.getLeftAncestors());
			}
		}
		public boolean hasNext() {
			return !nodes.isEmpty();
		}

		public BinaryNode next() {
			if (!hasNext()) throw new NoSuchElementException();
			BinaryNode temp = nodes.remove(0);
			if (temp.rightChild != null){
				nodes.addAll(0, temp.rightChild.getLeftAncestors());
			}
			return temp;
		}
	}
	
	// This is a pre-order iterator
	private class PreOrderIterator implements Iterator<BinarySearchTree<T>.BinaryNode> {
		private LinkedList<BinaryNode> nodes;

		public PreOrderIterator(){
			nodes = new LinkedList<BinaryNode>();
			if (root != null){
				nodes.addFirst(root);
			}
		}
		public boolean hasNext() {
			return !nodes.isEmpty();
		}

		public BinaryNode next() {
			if (!hasNext()) throw new NoSuchElementException();
			BinaryNode temp = nodes.removeFirst();
			if (temp.rightChild != null){
				nodes.addFirst(temp.rightChild);
			}
			if (temp.leftChild != null){
				nodes.addFirst(temp.leftChild);
			}
			return temp;
		}
	}
}
